{% extends "base.html" %}

{% block announce %}
    For updates follow <strong>@remilouf</strong> on
    <a href="https://twitter.com/squidfunk">
        <span class="twemoji twitter">
        {% include ".icons/fontawesome/brands/twitter.svg" %}
        </span>
        <strong>Twitter</strong>
    </a>
{% endblock %}
